package openio

import (
	"context"
	"fmt"
	"strconv"
	"strings"
	"time"

	"github.com/open-io/oio-exporter/common"
	"github.com/open-io/oio-exporter/daemon"
)

type openioMetaDaemonParams struct {
	Namespace     string
	ServiceType   string
	ServiceID     string
	ServiceIP     string
	ServicePort   int
	ServiceVolume string
	ProxyAddrs    []string
}

type openioMetaDaemon struct {
	daemon.BaseDaemon

	namespace     string
	serviceType   string
	serviceID     string
	serviceIP     string
	servicePort   int
	serviceVolume string
	serviceAddr   string
	proxyAddrs    []string

	proxyI int
}

type metaxInfoBody struct {
	Cache     map[string]int64
	Elections map[string]int64
}

func (d *openioMetaDaemon) Interval() time.Duration {
	return confInterval
}

func (d *openioMetaDaemon) Init(ctx context.Context) error {
	d.serviceAddr = fmt.Sprintf("%s:%d", d.serviceIP, d.servicePort)

	return nil
}

func (d *openioMetaDaemon) Collect(ctx context.Context, m *common.MetricData) error {
	serviceType := d.serviceType
	serviceAddr := d.serviceAddr
	proxyAddrs := d.proxyAddrs
	d.proxyI = (d.proxyI + 1) % len(proxyAddrs)
	proxyAddr := proxyAddrs[d.proxyI]

	url := fmt.Sprintf("http://%s/v3.0/forward/stats?id=%s", proxyAddr, serviceAddr)
	requestCtx, cancel := context.WithTimeout(ctx, confTimeout)
	defer cancel()

	counters, err := common.HTTPGetCsv(requestCtx, url, ' ')
	if err != nil {
		m.IncrError("openio", serviceType)
		return err
	}

	baseLabels := map[string]string{
		"namespace":    d.namespace,
		"service":      d.serviceID,
		"service_type": d.serviceType,
		"service_addr": serviceAddr,
		"volume":       d.serviceVolume,
	}

	for _, cntr := range counters {
		val, err := strconv.ParseFloat(cntr[2], 64)
		if err != nil {
			continue
		}
		if cntr[0] != "counter" {
			continue
		}
		parts := strings.Split(cntr[1], ".")
		if len(parts) == 2 {
			parts = append(parts, "*")
		}
		if len(parts) != 3 {
			continue
		}

		labels := common.CopyLabels(baseLabels)
		if parts[1] == "hits" || parts[1] == "time" {
			if parts[0] == "req" {
				labels["method"] = parts[2]
			} else { // "rep"
				labels["code"] = parts[2]
			}
		}

		m.Set(&common.Metric{
			Name:   parts[0] + "_" + parts[1],
			Labels: labels,
			Value:  val,
			TTL:    confInterval * 2,
		})
	}

	url = fmt.Sprintf("http://%s/v3.0/forward/info?id=%s", proxyAddr, serviceAddr)
	requestCtx, cancel = context.WithTimeout(ctx, confTimeout)
	defer cancel()

	svcInfo := metaxInfoBody{}
	err = common.HTTPGetJson(requestCtx, url, &svcInfo)
	if err != nil {
		m.IncrError("openio", serviceType)
		return err
	}

	allMetrics := map[string]map[string]int64{
		"base_cache_size/type":  svcInfo.Cache,
		"base_elections/status": svcInfo.Elections,
	}
	for t, metrics := range allMetrics {
		vars := strings.Split(t, "/")
		if len(vars) != 2 {
			continue
		}

		for dim, val := range metrics {
			labels := common.CopyLabels(baseLabels)
			labels[vars[1]] = dim
			m.Set(&common.Metric{
				Name:   vars[0],
				Labels: labels,
				Value:  float64(val),
				TTL:    confInterval * 2,
			})
		}
	}

	return nil
}
